from pandac.PandaModules import *
from direct.showbase import DirectObject

import math


class Mouse(DirectObject.DirectObject):
  def __init__(self,policeMan):
    self.policeMan = policeMan
    
    # We need to pick up on mouse events...
    self.accept('mouse1',self.recordStart,[False])
    self.accept('mouse1-up',self.select)
    self.accept('mouse3',self.recordStart,[True])
    self.accept('mouse3-up',self.order)

    base.camLens.setNearFar(0.1,200.0)

    self.pointThres = 0.025

    # Create box used for selection window drawing...
    self.sel = loader.loadModel('data/plane')
    self.sel.reparentTo(render2d)
    self.sel.setColor(Vec4(0.0,0.0,0.0,0.25))
    self.sel.setTransparency(TransparencyAttrib.MAlpha)
    self.sel.setLightOff()
    self.sel.hide()

    # Runtime variables...
    self.startX = 0.0
    self.startY = 0.0
    self.endX = 0.0
    self.endY = 0.0
    self.lmb = False
    self.rmb = False
    
    # Create task that manages selection info drawing...
    def selVis(task):
      self.sel.hide()
      if base.mouseWatcherNode.hasMouse():
        mPos = base.mouseWatcherNode.getMouse()
        self.curX = mPos[0]
        self.curY = mPos[1]
        
        if self.lmb:
          if abs(self.startX-self.curX)>self.pointThres or abs(self.startY-self.curY)>self.pointThres:
            self.sel.show()
            self.sel.setSx(0.5*abs(self.startX-self.curX))
            self.sel.setSz(0.5*abs(self.startY-self.curY))
            self.sel.setR(0.0)
            self.sel.setX(0.5*(self.startX+self.curX))
            self.sel.setZ(0.5*(self.startY+self.curY))
        elif self.rmb:
          length = math.sqrt((self.startX-self.curX)**2.0 + (self.startY-self.curY)**2.0)
          if length>self.pointThres:
            self.sel.show()
            self.sel.setSx(0.5*length)
            self.sel.setSz(0.5*self.pointThres)
            if self.curX>self.startX:
              ang = math.asin((self.startY-self.curY)/length)
            else:
              ang = -math.asin((self.startY-self.curY)/length)
            self.sel.setR(ang*180.0/math.pi)
            self.sel.setX(0.5*(self.startX+self.curX))
            self.sel.setZ(0.5*(self.startY+self.curY))
          
      return task.cont

    taskMgr.add(selVis,'selVis')
    

  def recordStart(self,which):
    mPos = base.mouseWatcherNode.getMouse()
    self.startX = mPos[0]
    self.startY = mPos[1]

    if which:
      self.rmb = True
    else:
      self.lmb = True

  def posToPlane(self,x,y):
    """Helper method - given a screen coordinate calculates where on the z=0.5 plane the mouse ray intercepts. As a (x,y) tuple. Returns None if there is a problem. (z=0.5 plane as the bobbles bottoms are 0.5m radius balls, so this is instinctivly where the player clicks.)"""
    near = Point3()
    far = Point3()

    base.camLens.extrude(Point2(x,y),near,far)
    near = render.getRelativePoint(base.camera,near)
    far = render.getRelativePoint(base.camera,far)

    far -= near
    far /= far[2]
    return (near[0] - far[0]*(near[2]-0.5),near[1] - far[1]*(near[2]-0.5))

  def select(self):
    self.lmb = False
    if base.mouseWatcherNode.hasMouse():
      mPos = base.mouseWatcherNode.getMouse()
      self.endX = mPos[0]
      self.endY = mPos[1]

      if abs(self.startX-self.endX)<self.pointThres and abs(self.startY-self.endY)<self.pointThres:
        self.policeMan.selectPoint(self.posToPlane(0.5*(self.startX+self.endX), 0.5*(self.startY+self.endY)))
      else:
        # We effectivly have 4 points, that define an arbitrary polygon once projected to the ground plane - this polygon needs to be sent to the selection code so it can find the relevant policeman...
        poly = []
        poly.append(self.posToPlane(self.startX,self.startY))
        poly.append(self.posToPlane(self.startX,self.endY))
        poly.append(self.posToPlane(self.endX,self.endY))
        poly.append(self.posToPlane(self.endX,self.startY))

        # Send to policeman manager, so it can do a selection...
        self.policeMan.selectRegion(poly)


  def order(self):
    self.rmb = False
    if base.mouseWatcherNode.hasMouse():
      mPos = base.mouseWatcherNode.getMouse()
      self.endX = mPos[0]
      self.endY = mPos[1]

      a = self.posToPlane(self.startX,self.startY)
      b = self.posToPlane(self.endX,self.endY)
    
      self.policeMan.move(a,b)