from bobbles import *

import random

# Parameters for the protestors...
minLeaveTime = 240.0
maxLeaveTime = 960.0
minWaterTime = 60.0
maxWaterTime = 240.0
minFoodTime = 120.0
maxFoodTime = 480.0
minToiletTime = 60.0
maxToiletTime = 180.0

normalCol = (196.0/255.0,132.0/255.0,130.0/255.0)
leaveCol = (95.0/255.0,155.0/255.0,157.0/255.0)
waterCol = (219.0/255.0,153.0/255.0,88.0/255.0)
foodCol = (116.0/255.0,200.0/255.0,111.0/255.0)
toiletCol = (212.0/255.0,210.0/255.0,73.0/255.0)


class ProtestorAI(BobbleAI):
  def __init__(self,pos):
    BobbleAI.__init__(self,pos)

    self.standTime = 0.0 # How long since the protestor saw significant movement, or neg during a random walk.
    self.task = 'protest'
    self.setStrength(random.uniform(0.8,1.2),random.uniform(0.8,1.2))
    self.setSpeed(random.uniform(0.8,1.2),random.uniform(0.8,1.2))

    self.time = 0.0 # Time since the protestor last did any serious thinking.

    # Generate countdowns till certain behaviour...
    self.nextLeave = random.uniform(minLeaveTime,maxLeaveTime)
    self.nextWater = random.uniform(minWaterTime,maxWaterTime)
    self.nextFood = random.uniform(minFoodTime,maxFoodTime)
    self.nextToilet = random.uniform(minToiletTime,maxToiletTime)
    
    # Generate variables indicate unusual behaviour...
    self.noJob = random.random()<0.1 # Never leaves
    self.waterBottle = random.random()<0.5 # Need no water
    self.packedLunch = random.random()<0.25 # Need no food
    self.wearsNappy = random.random()<0.01 # Doesn't use the toilet.

    # This defines how safe the bobble is with regards to leaving enough time to get places...
    self.safety = random.uniform(2.0,5.0)


  def step(self,body):
    # Safety check - if its fallen out of the level destroy...
    if body.getPosition()[2]<-100.0:
      self.die()
      if self.manager.pList.has_key(id(self)): del self.manager.pList[id(self)]

    if self.standTime < 0.0:
      # We are currently being evasive to avoid an obstruction...
      self.standTime += globalClock.getDt()
      if not (self.standTime < 0.0):
        dPos = self.dest.getRandomHalfLoc()
        self.setDest(dPos[0],dPos[1])
    else:
      # Check if we have a significant velocity - if not incriment the not moved counter...
      if body.getLinearVel().length()<0.1:
        self.standTime += globalClock.getDt()
      else:
        self.standTime = 0.0

      # If we have been standing around too long flip to problem fixing mode - go in a random direction for a bit, or choose a random node next to the destination node to go to...
      if self.standTime>0.5:
        if random.random()<0.75:
          self.standTime = random.uniform(-8.0,-4.0)

          # Choose a random direction to head in...
          pos = body.getPosition()
          a = random.random()*math.pi*2.0
          self.setDest(pos[0] + math.sin(a)*100.0,pos[1] + math.cos(a)*100.0)
        else:
          self.standTime = 0.0
          self.dest = self.dest.getNear()
          dPos = self.dest.getRandomHalfLoc()
          self.setDest(dPos[0],dPos[1])

    # Incriment the time since last though counter...
    self.time += globalClock.getDt()
    
    # Check if we are at our target destination...
    if self.dest.getDist(body) < self.dest.getRadius():
      # Update the countdowns...
      self.nextLeave -= self.time
      self.nextWater -= self.time
      self.nextFood -= self.time
      self.nextToilet -= self.time
      self.time = 0.0

      # Refresh countdowns depending on what the location provides...
      leaveDist = self.dest.distanceTo('exit')
      waterDist = self.dest.distanceTo('water')
      foodDist = self.dest.distanceTo('food')
      toiletDist = self.dest.distanceTo('toilet')
      
      if waterDist<1.0:
        self.nextWater = random.uniform(minWaterTime,maxWaterTime)
        if self.task=='water':
          self.task = 'protest'
          self.setCol(normalCol)
      if foodDist<1.0:
        self.nextFood = random.uniform(minFoodTime,maxFoodTime)
        if self.task=='food':
          self.task = 'protest'
          self.setCol(normalCol)
      if toiletDist<1.0:
        self.nextToilet = random.uniform(minToiletTime,maxToiletTime)
        if self.task=='toilet':
          self.task = 'protest'
          self.setCol(normalCol)

      if self.task=='leave' and leaveDist<1.0: # If leaving and can leave then do so.
        self.die()
        if self.manager.pList.has_key(id(self)): del self.manager.pList[id(self)]
      
      # React to any zeroed countdowns...
      if not self.waterBottle and self.nextWater<0.0:
        self.die()
        if self.manager.pList.has_key(id(self)): del self.manager.pList[id(self)]
        print 'dehydration'
        
      if not self.packedLunch and self.nextFood<0.0:
        self.nextFood = 10000.0
        if self.noJob:
          print 'starved to death'
          self.die()
          if self.manager.pList.has_key(id(self)): del self.manager.pList[id(self)]
        else:
          print 'hungry'
          self.task = 'leave'
          self.setCol(leaveCol)
        
      if not self.wearsNappy and self.nextToilet<0.0:
        self.nextToilet = random.uniform(minToiletTime,maxToiletTime)
        if self.task=='toilet':
          self.task = 'protest'
          self.setCol(normalCol)
          print 'defecation on the street'
        else:
          print 'clean underwear please'
          self.task = 'leave'
          self.setCol(leaveCol)


      # Choose mode - head to things we need to so will get there in time - safety threshold varies by bobble...
      if self.task=='protest': # Choice only if protesting - anything else and we continue with that.
        if not self.noJob and (self.safety*leaveDist/self.targSpeed > self.nextLeave):
          self.task = 'leave'
          self.setCol(leaveCol)
        elif not self.waterBottle and (self.safety*waterDist/self.targSpeed > self.nextWater):
          self.task = 'water'
          self.setCol(waterCol)
        elif not self.wearsNappy and (self.safety*toiletDist/self.targSpeed > self.nextToilet):
          self.task = 'toilet'
          self.setCol(toiletCol)
        elif not self.packedLunch and (self.safety*foodDist/self.targSpeed > self.nextFood):
          self.task = 'food'
          self.setCol(foodCol)
      
      # As we have reached our destination we have to decide on the next - this depends on mode...
      if self.task=='protest': self.dest = self.dest.getMarch()
      elif self.task=='leave': self.dest = self.dest.nodeTo('exit')
      elif self.task=='water': self.dest = self.dest.nodeTo('water')
      elif self.task=='food': self.dest = self.dest.nodeTo('food')
      elif self.task=='toilet': self.dest = self.dest.nodeTo('toilet')

      # Update the bobbleAI's target position...
      dPos = self.dest.getRandomHalfLoc()
      self.setDest(dPos[0],dPos[1])
    
    return BobbleAI.step(self,body)



class ProtestorManager:
  def __init__(self,bobbles,routing):
    self.bobbles = bobbles
    self.routing = routing

    self.pList = dict()

    # Target state...
    self.pCount = 128 # How many protestors we want.
    self.pGrow = 5 # Maximum number of new protestors per second.

    # Create management task...
    def protestManager(task):
      #print len(self.pList), 'protestors'

      grow = random.choice(range(self.pGrow+1))
      for i in xrange(max(min(grow,self.pCount-len(self.pList)),0)):
        self.addProtestor()
      
      return task.again

    taskMgr.doMethodLater(1.0,protestManager,'protestManager')

  def addProtestor(self):
    # Select a random position and rotation...
    node = self.routing.randomSpecial('enter')
    pos = node.getRandomHalfLoc()
    h = 0.0

    # Create the bobble, store in db...
    np = self.bobbles.addBobble(pos,h,normalCol,ProtestorAI)
    self.pList[id(np)] = np

    # Prep it for its protest...
    np.dest = node
    np.manager = self