from pandac.PandaModules import *

import math
import random


class Location:
  def __init__(self,node):
    self.name = node.getTag('Name')
    self.radius = float(node.getTag('Radius'))
    self.near = filter(lambda x: x!='',node.getTag('Near').split(','))
    self.march = filter(lambda x: x!='',node.getTag('March').split(','))
    self.special = filter(lambda x: x!='',node.getTag('Special').split(','))
    self.x = node.getX()
    self.y = node.getY()
    self.z = node.getZ()

  def getName(self):
    return self.name

  def getRadius(self):
    return self.radius

  def getX(self):
    return self.x

  def getY(self):
    return self.y

  def getRandomHalfLoc(self):
    """Returns a random Vec3 that describes a position in the half radius circle around the node. Uses rejection sampling."""
    while True:
      x = random.uniform(-1.0,1.0)
      y = random.uniform(-1.0,1.0)
      if x**2.0 + y**2.0 < 1.0:
        return Vec3(self.x + self.radius*0.5*x,self.y + self.radius*0.5*y,self.z)

  def getDist(self,body):
    """Returns the distance in the x,y plane from body to the node."""
    pos = body.getPosition()
    xd = pos[0] - self.x
    yd = pos[1] - self.y
    return math.sqrt(xd**2.0 + yd**2.0)

  def getMarch(self):
    """Returns the next node if following the march."""
    return random.choice(self.march)[0]

  def getNear(self):
    """Returns a random adjacent node."""
    return random.choice(self.near)[0]

  def distanceTo(self,special):
    """Distance to a special from this node."""
    return self.route[special][1]

  def nodeTo(self,special):
    return self.route[special][0]

  def __str__(self):
    return self.name



class Routing:
  """Loads the routing nodes and prepares the data required for the boobles to have a good protest."""
  def __init__(self):
    # Load the locations data...
    locs = loader.loadModel('data/locations')

    # Get all the nodes, build the basic structure...
    self.node = dict()
    nodes = locs.findAllMatches('**/=IsA=Location')
    for i in xrange(nodes.size()):
      loc = Location(nodes[i])
      self.node[loc.getName()] = loc

    # Convert the list of near nodes and the list of march nodes into a list of tuples (node,distance)...
    for name,loc in self.node.iteritems():
      def nameToTup(name):
        node = self.node[name]
        dist = math.sqrt((loc.x - node.x)**2.0 + (loc.y - node.y)**2.0)
        return (node,dist)
      loc.near = map(nameToTup,loc.near)
      loc.march = map(nameToTup,loc.march)
      loc.offMarch = len(loc.march)==0
        

    # Finish off the march routes - make 'em march to the march!...
    while True:
      change = False
      for name,loc in self.node.iteritems():
        if loc.offMarch:
          for near,dist in loc.near:
            # Get the route, if it exists...
            if near.offMarch:
              if len(near.route)==1:
                r = (near,dist + near.route[0][1])
              else:
                r = None
            else:
              r = (near,dist)

            # If the route exists see if its better than the current, if it exists...
            if r!=None:
              if len(loc.march)==1:
                if loc.march[0][1]>r[1]:
                  loc.march = [r]
                  change = True
              else:
                loc.march = [r]
                change = True
      if change==False:
        break

    # Prepare the route database - best route for each resource, also make lists of which nodes have which attributes...
    self.bySpecial = dict()
    for name,loc in self.node.iteritems():
      loc.route = dict()
      for special in loc.special:
        loc.route[special] = (loc,0.0)
        if self.bySpecial.has_key(special):
          self.bySpecial[special].append(loc)
        else:
          self.bySpecial[special] = [loc]

    # Now, for every node, calculate the shortest route to each of the resources found in special...
    while True:
      changed = False
      for name,loc in self.node.iteritems():
        # See if we can improve/add to the routes for this node - iterate all near nodes and their routes...
        for near,dist in loc.near:
          mult = 1.0
          if loc in map(lambda x:x[0],near.march):
            if loc.offMarch or near.offMarch:
              mult = 1.5
            else:
               mult = 3.0 # Going again the flow bad.

          for special,locDist in near.route.iteritems():
            if loc.route.has_key(special):
              d = dist + mult*locDist[1]
              if d < loc.route[special][1]:
                loc.route[special] = (near,d)
                changed = True
            else:
              loc.route[special] = (near,dist + mult*locDist[1])
              changed = True
      if changed==False:
        break


  def randomSpecial(self,name):
    """Returns a random node that has the given name; crashes if it doesn't exist."""
    return random.choice(self.bySpecial[name])